# Get version
import re

name = "chromatinhd"

import setuptools

VERSIONFILE = "src/" + name + "/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    version = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

# get long description
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name=name,
    version=version,
    author="Wouter Saelens",
    author_email="wouter.saelens@gmail.com",
    description="Modeling of chromatin + transcriptomics data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DeplanckeLab/ChromatinHD",
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    install_requires=[
        "torch",  # --extra-index-url https://download.pytorch.org/whl/cu113
        "torch-scatter",
        # "torch==1.12.1",  # --extra-index-url https://download.pytorch.org/whl/cu113
        # "torch-scatter",  # --find-links https://data.pyg.org/whl/torch-1.12.1+cu113.html
        "scanpy",
        "matplotlib",
        "numpy",
        "seaborn",
        "Cython",
        "fisher",
        "diskcache",
        "appdirs",
        "xarray",
        "pysam",
    ],
    extras_require={
        "full": [],
        "dev": [
            "pre-commit",
            "pytest",
            "coverage",
            "black",
            "pylint",
            "jupytext",
            "pytest",
            "statsmodels",
            "mkdocs",
            "mkdocs-material",
            "mkdocstrings[python]",
            "mkdocs-jupyter",
            "mike",
            "cairosvg",  # for mkdocs social
            "pillow",  # for mkdocs social
            # "faiss-cpu",
        ],
        "eqtl": ["cyvcf2", "xarray"],
    },
)



from setuptools import setup
from Cython.Build import cythonize
import numpy as np

setup(
    name='Extract motifs',
    ext_modules=cythonize("extract_motifs.pyx"),
    include_dirs=[np.get_include()],
    zip_safe=False,
)

setup(
    name='Extract fragments',
    ext_modules=cythonize("extract_fragments.pyx"),
    include_dirs=[np.get_include()],
    zip_safe=False,
)


import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import chromatinhd


def center(coords, region):
    if coords.shape[0] == 0:
        coords = pd.DataFrame(columns=["start", "end", "method"])
    else:
        coords[["start", "end"]] = [
            [
                (peak["start"] - region["tss"]) * region["strand"],
                (peak["end"] - region["tss"]) * region["strand"],
            ][:: region["strand"]]
            for _, peak in coords.iterrows()
        ]
    return coords


def get_genes_plotdata(region, genome="GRCh38"):
    biomart_dataset = chd.biomart.Dataset.from_genome(genome)
    if "chrom" not in region.index:
        region["chrom"] = region["chr"]
    canonical_transcripts = chd.biomart.get_canonical_transcripts(
        biomart_dataset, chrom=region["chrom"], start=region["start"], end=region["end"]
    )
    exons = chd.biomart.get_exons(
        biomart_dataset, chrom=region["chrom"], start=region["start"], end=region["end"]
    )

    plotdata_genes = canonical_transcripts
    plotdata_genes = center(plotdata_genes, region)

    plotdata_exons = exons.rename(
        columns={
            "exon_chrom_start": "start",
            "exon_chrom_end": "end",
            "ensembl_gene_id": "gene",
        }
    )
    plotdata_exons = center(plotdata_exons, region)

    plotdata_coding = exons.dropna().rename(
        columns={
            "genomic_coding_start": "start",
            "genomic_coding_end": "end",
            "ensembl_gene_id": "gene",
        }
    )
    plotdata_coding = center(plotdata_coding, region)

    return plotdata_genes, plotdata_exons, plotdata_coding


class Genes(chromatinhd.grid.Ax):
    def __init__(
        self,
        plotdata_genes,
        plotdata_exons,
        plotdata_coding,
        gene_id,
        region,
        window,
        width,
        full_ticks=False,
        label_genome=False,
        symbol=None,
    ):
        super().__init__((width, len(plotdata_genes) * 0.08 + 0.01))

        ax = self.ax

        ax.xaxis.tick_top()
        if label_genome:
            if symbol is None:
                symbol = gene_id
            ax.set_xlabel("Distance to $\\mathit{" + symbol + "}$ TSS")
        ax.xaxis.set_label_position("top")
        ax.tick_params(axis="x", length=2, pad=0, labelsize=8, width=0.5)
        ax.xaxis.set_major_formatter(chromatinhd.plot.gene_ticker)

        sns.despine(ax=ax, right=True, left=True, bottom=True, top=True)

        ax.set_xlim(*window)

        ax.set_yticks([])
        ax.set_ylabel("")

        if len(plotdata_genes) == 0:
            return

        plotdata_genes["ix"] = np.arange(len(plotdata_genes))

        ax.set_ylim(-0.5, plotdata_genes["ix"].max() + 0.5)
        if full_ticks:
            ax.set_xticks(np.arange(window[0], window[1] + 1, 500))
            ax.xaxis.set_major_formatter(mpl.ticker.ScalarFormatter())
            ax.tick_params(
                axis="x",
                length=2,
                pad=0,
                labelsize=8,
                width=0.5,
                labelrotation=90,
            )
        for gene, gene_info in plotdata_genes.iterrows():
            y = gene_info["ix"]
            is_oi = gene == gene_id
            ax.plot(
                [gene_info["start"], gene_info["end"]],
                [y, y],
                color="black" if is_oi else "grey",
            )

            if pd.isnull(gene_info["symbol"]):
                symbol = gene_info.name
            else:
                symbol = gene_info["symbol"]
            strand = gene_info["strand"] * region["strand"]
            if (gene_info["start"] > window[0]) & (gene_info["start"] < window[1]):
                label = symbol + " → " if strand == 1 else " ← " + symbol
                ha = "right"
                # ha = "left" if (strand == -1) else "right"

                ax.text(
                    gene_info["start"],
                    y,
                    label,
                    style="italic",
                    ha=ha,
                    va="center",
                    fontsize=6,
                    weight="bold" if is_oi else "regular",
                )
            elif (gene_info["end"] > window[0]) & (gene_info["end"] < window[1]):

                label = " → " + symbol if strand == 1 else symbol + " ← "
                ha = "left"

                ax.text(
                    gene_info["end"],
                    y,
                    label,
                    style="italic",
                    ha=ha,
                    va="center",
                    fontsize=6,
                    weight="bold" if is_oi else "regular",
                )
            else:
                ax.text(
                    0,
                    y,
                    "(" + symbol + ")",
                    style="italic",
                    ha="center",
                    va="center",
                    fontsize=6,
                    bbox=dict(facecolor="#FFFFFF88", boxstyle="square,pad=0", lw=0),
                )

            plotdata_exons_gene = plotdata_exons.query("gene == @gene")
            h = 1
            for exon, exon_info in plotdata_exons_gene.iterrows():
                rect = mpl.patches.Rectangle(
                    (exon_info["start"], y - h / 2),
                    exon_info["end"] - exon_info["start"],
                    h,
                    fc="white",
                    ec="#333333",
                    lw=1.0,
                    zorder=9,
                )
                ax.add_patch(rect)

            plotdata_coding_gene = plotdata_coding.query("gene == @gene")
            for coding, coding_info in plotdata_coding_gene.iterrows():
                rect = mpl.patches.Rectangle(
                    (coding_info["start"], y - h / 2),
                    coding_info["end"] - coding_info["start"],
                    h,
                    fc="#333333",
                    ec="#333333",
                    lw=1.0,
                    zorder=10,
                )
                ax.add_patch(rect)

        # vline at tss
        ax.axvline(0, color="#888888", lw=0.5, zorder=-1, dashes=(2, 2))

    @classmethod
    def from_region(cls, region, genome="GRCh38", window=None, **kwargs):
        if window is None:
            assert "tss" in region
            window = np.array(
                [region["start"] - region["tss"], region["end"] - region["tss"]]
            )
        plotdata_genes, plotdata_exons, plotdata_coding = get_genes_plotdata(
            region, genome=genome
        )

        return cls(
            plotdata_genes=plotdata_genes,
            plotdata_exons=plotdata_exons,
            plotdata_coding=plotdata_coding,
            region=region,
            gene_id=region.name,
            window=window,
            **kwargs,
        )


import chromatinhd as chd
import pandas as pd
import numpy as np


from chromatinhd.grid.broken import Broken, Panel
from chromatinhd.plot import gene_ticker
import matplotlib as mpl
import seaborn as sns


def filter_start_end(x, start, end):
    y = x.loc[~((x["end"] < start) | (x["start"] > end))]
    return y


def filter_position(x, start, end):
    y = x.loc[~((x["position"] < start) | (x["position"] > end))]
    return y


class GenesBrokenBase(Broken):
    def __init__(
        self,
        plotdata_genes,
        plotdata_exons,
        plotdata_coding,
        regions,
        gene_id,
        promoter,
        window,
        width,
        gap,
        full_ticks=False,
        *args,
        **kwargs,
    ):
        height = len(plotdata_genes) * 0.08
        super().__init__(
            regions=regions,
            height=height,
            width=width,
            gap=gap,
            *args,
            **kwargs,
        )

        ylim = (-0.5, plotdata_genes["ix"].max() + 0.5)

        for ((region, region_info), (panel, ax)) in zip(
            regions.iterrows(), self.elements[0]
        ):
            ax.xaxis.tick_top()
            ax.set_yticks([])
            ax.set_ylabel("")
            # ax.set_xlabel("Distance to TSS")
            ax.xaxis.set_label_position("top")
            ax.tick_params(axis="x", length=2, pad=0, labelsize=8, width=0.5)
            ax.xaxis.set_major_formatter(gene_ticker)

            for spine in ax.spines.values():
                spine.set_visible(False)

            ax.set_xlim(region_info["start"], region_info["end"])
            ax.set_ylim(*ylim)

            plotdata_genes_region = filter_start_end(
                plotdata_genes, region_info["start"], region_info["end"]
            )

            for gene, gene_info in plotdata_genes_region.iterrows():
                y = gene_info["ix"]
                is_oi = gene == gene_id
                ax.plot(
                    [gene_info["start"], gene_info["end"]],
                    [y, y],
                    color="black" if is_oi else "grey",
                )

                plotdata_exons_gene = plotdata_exons.query("gene == @gene")
                plotdata_exons_gene = filter_start_end(
                    plotdata_exons_gene, region_info["start"], region_info["end"]
                )
                h = 1
                for exon, exon_info in plotdata_exons_gene.iterrows():
                    rect = mpl.patches.Rectangle(
                        (exon_info["start"], y - h / 2),
                        exon_info["end"] - exon_info["start"],
                        h,
                        fc="white",
                        ec="#333333",
                        lw=1.0,
                        zorder=9,
                    )
                    ax.add_patch(rect)

                plotdata_coding_gene = plotdata_coding.query("gene == @gene")
                plotdata_coding_gene = filter_start_end(
                    plotdata_coding_gene, region_info["start"], region_info["end"]
                )
                for coding, coding_info in plotdata_coding_gene.iterrows():
                    rect = mpl.patches.Rectangle(
                        (coding_info["start"], y - h / 2),
                        coding_info["end"] - coding_info["start"],
                        h,
                        fc="#333333",
                        ec="#333333",
                        lw=1.0,
                        zorder=10,
                    )
                    ax.add_patch(rect)

            # vline at tss
            ax.axvline(0, color="#888888", lw=0.5, zorder=-1, dashes=(2, 2))

        ax = self[0, 0].ax
        ax.set_yticks(np.arange(len(plotdata_genes)))
        ax.set_yticklabels(plotdata_genes["symbol"], fontsize=6, style="italic")
        ax.tick_params(axis="y", length=0, pad=2, width=0.5)


class GenesBroken(GenesBrokenBase):
    def __init__(self, promoter, genome_folder, window, *args, **kwargs):
        plotdata_genes, plotdata_exons, plotdata_coding = get_genes_plotdata(
            promoter, genome_folder, window
        )

        return super().__init__(
            *args,
            plotdata_genes=plotdata_genes,
            plotdata_exons=plotdata_exons,
            plotdata_coding=plotdata_coding,
            promoter=promoter,
            gene_id=promoter.name,
            window=window,
            **kwargs,
        )



import chromatinhd.data.transcriptome
import dataclasses
import torch

from .fragments import Fragments
from .minibatches import Minibatch
from .transcriptome import Transcriptome


@dataclasses.dataclass
class Result:
    transcriptome: Transcriptome
    fragments: Fragments
    minibatch: Minibatch

    def to(self, device):
        self.transcriptome.to(device)
        self.fragments.to(device)
        self.minibatch.to(device)
        return self


class TranscriptomeFragments:
    def __init__(
        self,
        fragments: chromatinhd.data.fragments.Fragments,
        transcriptome: chromatinhd.data.transcriptome.Transcriptome,
        cellxgene_batch_size: int,
        layer: str = "X",
    ):
        self.fragments = Fragments(fragments, cellxgene_batch_size=cellxgene_batch_size)
        self.transcriptome = Transcriptome(transcriptome, layer=layer)

    def load(self, minibatch):
        return Result(
            transcriptome=self.transcriptome.load(minibatch),
            fragments=self.fragments.load(minibatch),
            minibatch=minibatch,
        )



import chromatinhd
import chromatinhd.grid
import matplotlib as mpl
import seaborn as sns
import numpy as np
import pandas as pd


def get_cmap_atac_diff():
    return mpl.cm.RdBu_r


def get_norm_atac_diff():
    return mpl.colors.Normalize(np.log(1 / 4), np.log(4.0), clip=True)

class Differential(chromatinhd.grid.Wrap):
    def __init__(
        self,
        plotdata,
        plotdata_mean,
        cluster_info,
        width,
        window=None,
        panel_height=0.5,
        plotdata_empirical=None,
        show_atac_diff=True,
        cmap_atac_diff=mpl.cm.RdBu_r,
        norm_atac_diff=mpl.colors.Normalize(np.log(1 / 4), np.log(4.0), clip=True),
        ymax=20,
        **kwargs,
    ):
        super().__init__(ncol=1, **{"padding_height": 0, **kwargs})
        self.show_atac_diff = show_atac_diff
        self.cmap_atac_diff = cmap_atac_diff
        self.norm_atac_diff = norm_atac_diff
        self.window = window
        self.cluster_info = cluster_info

        # check plotdata
        plotdata = (
            plotdata.reset_index()
            .assign(coord=lambda x: x.coord.astype(int))
            .set_index(["cluster", "coord"])
        )
        plotdata_mean = (
            plotdata_mean.reset_index()
            .assign(coord=lambda x: x.coord.astype(int))
            .set_index(["coord"])
        )

        if "prob_diff" not in plotdata.columns:
            plotdata["prob_diff"] = plotdata["prob"] - plotdata_mean["prob"]

        if window is None:
            window = (
                plotdata.index.get_level_values("coord").min(),
                plotdata.index.get_level_values("coord").max(),
            )

        for cluster, cluster_info_oi in self.cluster_info.iterrows():
            panel, ax = chromatinhd.grid.Ax((width, panel_height))
            self.add(panel)

            ax.set_ylim(0, ymax)
            ax.set_xlim(*window)
            ax.axvline(0, dashes=(1, 1), color="#AAA", zorder=-1)

            text = ax.annotate(
                text=f"{cluster_info_oi['label']}",
                xy=(0, 1),
                xytext=(2, -2),
                textcoords="offset points",
                xycoords="axes fraction",
                ha="left",
                va="top",
                fontsize=10,
                color="#333",
            )
            text.set_path_effects(
                [
                    mpl.patheffects.Stroke(linewidth=2, foreground="white"),
                    mpl.patheffects.Normal(),
                ]
            )

            if plotdata_empirical is not None:
                # empirical distribution of atac-seq cuts
                plotdata_empirical_cluster = plotdata_empirical.query(
                    "cluster == @cluster"
                )
                ax.fill_between(
                    plotdata_empirical_cluster["coord"],
                    np.exp(plotdata_empirical_cluster["prob"]),
                    alpha=0.2,
                    color="#333",
                )

            ax.set_xticks([])

            if cluster != self.cluster_info.index[0]:
                ax.set_yticks([])
        ax.annotate(
            text="Mean &\ndifferential\naccessibility",
            xy=(0, len(cluster_info) / 2),
            xycoords="axes fraction",
            xytext=(-20, 0),
            textcoords="offset points",
            rotation=0,
            ha="right",
            va="center",
        )

        self.draw(plotdata, plotdata_mean)

    def draw(self, plotdata, plotdata_mean):
        self.artists = []

        for ax, cluster in zip(self.elements, self.cluster_info.index):
            ax = ax.ax
            if self.show_atac_diff:
                # posterior distribution of atac-seq cuts
                plotdata_cluster = plotdata.xs(cluster, level="cluster")
                (background,) = ax.plot(
                    plotdata_mean.index,
                    np.exp(plotdata_mean["prob"]),
                    color="black",
                    lw=1,
                    zorder=1,
                    linestyle="dashed",
                )
                (differential,) = ax.plot(
                    plotdata_cluster.index,
                    np.exp(plotdata_cluster["prob"]),
                    color="black",
                    lw=1,
                    zorder=1,
                )
                polygon = ax.fill_between(
                    plotdata_mean.index,
                    np.exp(plotdata_mean["prob"]),
                    np.exp(plotdata_cluster["prob"]),
                    color="black",
                    zorder=0,
                )

                # up/down gradient
                verts = np.vstack([p.vertices for p in polygon.get_paths()])
                c = plotdata_cluster["prob_diff"].values
                c[c == np.inf] = 0.0
                c[c == -np.inf] = -10.0
                gradient = ax.imshow(
                    c.reshape(1, -1),
                    cmap=self.cmap_atac_diff,
                    aspect="auto",
                    extent=[
                        verts[:, 0].min(),
                        verts[:, 0].max(),
                        verts[:, 1].min(),
                        verts[:, 1].max(),
                    ],
                    zorder=25,
                    norm=self.norm_atac_diff,
                )
                gradient.set_clip_path(polygon.get_paths()[0], transform=ax.transData)
                polygon.set_alpha(0)

                self.artists.extend([gradient, polygon, background, differential])

    def get_artists(self):
        return self.artists



import math

import torch
from torch.nn import functional as F
import numpy as np


def unconstrained_linear_spline(
    inputs, unnormalized_pdf, inverse=False, tail_bound=1.0, tails="linear"
):
    inside_interval_mask = (inputs >= -tail_bound) & (inputs <= tail_bound)
    outside_interval_mask = ~inside_interval_mask

    outputs = torch.zeros_like(inputs)
    logabsdet = torch.zeros_like(inputs)

    if tails == "linear":
        outputs[outside_interval_mask] = inputs[outside_interval_mask]
        logabsdet[outside_interval_mask] = 0
    else:
        raise RuntimeError("{} tails are not implemented.".format(tails))

    outputs[inside_interval_mask], logabsdet[inside_interval_mask] = linear_spline(
        inputs=inputs[inside_interval_mask],
        unnormalized_pdf=unnormalized_pdf[inside_interval_mask, :],
        inverse=inverse,
        left=-tail_bound,
        right=tail_bound,
        bottom=-tail_bound,
        top=tail_bound,
    )

    return outputs, logabsdet


def linear_spline(
    inputs, unnormalized_pdf, inverse=False, left=0.0, right=1.0, bottom=0.0, top=1.0
):
    """
    Reference:
    > Müller et al., Neural Importance Sampling, arXiv:1808.03856, 2018.
    """
    if not inverse and (torch.min(inputs) < left or torch.max(inputs) > right):
        raise ValueError("Inputs outside domain")
    elif inverse and (torch.min(inputs) < bottom or torch.max(inputs) > top):
        raise ValueError("Inputs outside domain")

    if inverse:
        inputs = (inputs - bottom) / (top - bottom)
    else:
        inputs = (inputs - left) / (right - left)

    num_bins = unnormalized_pdf.size(-1)

    pdf = F.softmax(unnormalized_pdf, dim=-1)

    cdf = torch.cumsum(pdf, dim=-1)
    cdf[..., -1] = 1.01
    cdf = F.pad(cdf, pad=(1, 0), mode="constant", value=0.0)

    if inverse:
        inv_bin_idx = torch.searchsorted(cdf, inputs, right=False)

        bin_boundaries = (
            torch.linspace(0, 1, num_bins + 1)
            .view([1] * inputs.dim() + [-1])
            .expand(*inputs.shape, -1)
        )

        slopes = (cdf[..., 1:] - cdf[..., :-1]) / (
            bin_boundaries[..., 1:] - bin_boundaries[..., :-1]
        )
        offsets = cdf[..., 1:] - slopes * bin_boundaries[..., 1:]

        inv_bin_idx = inv_bin_idx.unsqueeze(-1)
        input_slopes = slopes.gather(-1, inv_bin_idx)[..., 0]
        input_offsets = offsets.gather(-1, inv_bin_idx)[..., 0]

        outputs = (inputs - input_offsets) / input_slopes
        outputs = torch.clamp(outputs, 0, 1)

        logabsdet = -torch.log(input_slopes)
    else:
        bin_pos = inputs * num_bins

        bin_idx = torch.floor(bin_pos).long()
        bin_idx[bin_idx >= num_bins] = num_bins - 1

        alpha = bin_pos - bin_idx.float()

        if bin_idx.ndim < pdf.ndim:
            bin_idx = bin_idx.unsqueeze(-1)

        input_pdfs = pdf.gather(-1, bin_idx)  # [..., 0]

        outputs = cdf.gather(-1, bin_idx)  # [..., 0]
        outputs += alpha * input_pdfs
        outputs = torch.clamp(outputs, 0, 1)

        bin_width = 1.0 / num_bins
        logabsdet = torch.log(input_pdfs) - np.log(bin_width)

    if inverse:
        outputs = outputs * (right - left) + left
        logabsdet = logabsdet - math.log(top - bottom) + math.log(right - left)
    else:
        outputs = outputs * (top - bottom) + bottom
        logabsdet = logabsdet + math.log(top - bottom) - math.log(right - left)

    return outputs, logabsdet



from . import fragments
from . import transcriptome
from . import motifscan
from .fragments import Fragments
from .transcriptome import Transcriptome
from .genotype import Genotype
from .clustering import Clustering
from .motifscan import Motifscan, Motiftrack, GWAS
from .regions import Regions
from . import folds

__all__ = ["Fragments", "Transcriptome", "Regions", "folds"]



import tqdm.auto as tqdm
import torch
import numpy as np
from chromatinhd.train import Trace

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class Trainer:
    def __init__(
        self,
        model,
        loaders_train,
        loaders_validation,
        minibatcher_train,
        minibatcher_validation,
        optim,
        hooks_checkpoint=None,
        hooks_checkpoint2=None,
        device="cuda",
        n_epochs=30,
        checkpoint_every_epoch=1,
        optimize_every_step=1,
    ):
        self.model = model
        self.loaders_train = loaders_train
        self.loaders_validation = loaders_validation

        self.trace = Trace()

        self.optim = optim

        self.step_ix = 0
        self.epoch = 0
        self.n_epochs = n_epochs

        self.checkpoint_every_epoch = checkpoint_every_epoch
        self.optimize_every_step = optimize_every_step

        self.device = device

        self.minibatcher_train = minibatcher_train
        self.minibatcher_validation = minibatcher_validation

        self.hooks_checkpoint = hooks_checkpoint if hooks_checkpoint is not None else []
        self.hooks_checkpoint2 = (
            hooks_checkpoint2 if hooks_checkpoint2 is not None else []
        )

    def train(self):
        import gc

        gc.collect()
        torch.cuda.empty_cache()

        self.model = self.model.to(self.device)

        self.loaders_train.initialize(self.minibatcher_train)
        self.loaders_validation.initialize(self.minibatcher_validation)

        continue_training = True

        n_steps_total = self.n_epochs * len(self.loaders_train)
        pbar = tqdm.tqdm(total=n_steps_total, leave=False)

        while (self.epoch < self.n_epochs) and (continue_training):
            pbar.set_description(f"epoch {self.epoch}")

            # checkpoint if necessary
            if (self.epoch % self.checkpoint_every_epoch) == 0:
                for hook in self.hooks_checkpoint:
                    hook.start()

                with torch.no_grad():
                    for data_validation in self.loaders_validation:
                        data_validation = data_validation.to(self.device)

                        loss = self.model(data_validation).sum()

                        self.trace.append(
                            loss.item(), self.epoch, self.step_ix, "validation"
                        )

                        for hook in self.hooks_checkpoint:
                            hook.run_individual(self.model, data_validation)

                logger.info(f"{'•'} {self.epoch}/{self.n_epochs} {'step':>15}")
                self.trace.checkpoint(logger)

                for hook in self.hooks_checkpoint:
                    hook.finish()

                for hook in self.hooks_checkpoint2:
                    hook.run(self.model)

            # train
            for data_train in self.loaders_train:
                # actual training
                data_train = data_train.to(self.device)

                loss = self.model.forward(data_train).sum()
                loss.backward()

                # check if optimization
                if (self.step_ix % self.optimize_every_step) == 0:
                    self.optim.step()
                    self.optim.zero_grad()

                self.step_ix += 1
                pbar.update()

                self.trace.append(loss.item(), self.epoch, self.step_ix, "train")
            self.epoch += 1

        pbar.update(n_steps_total)
        pbar.close()

        self.model = self.model.to("cpu")




import matplotlib.ticker as ticker

import decimal

def count_zeros(value):
    decimal_value = str(decimal.Decimal((value)))
    if "." in decimal_value:
        count = len(decimal_value.split(".")[1])
    else:
        count = 0
    return count


def custom_gene_formatter(value, tick_pos):
    abs_value = int(abs(value))
    if abs_value >= 1000000:
        zeros = len(str(abs_value).rstrip("0")) - 1
        abs_value = abs_value / 1000000
        suffix = "mb"
    elif abs_value >= 1000:
        zeros = len(str(abs_value).rstrip("0")) - 1
        abs_value = abs_value / 1000
        suffix = "kb"
    elif abs_value == 0:
        zeros = 0
        return "TSS"
    else:
        zeros = 0
        suffix = "b"

    formatted_value = ("{abs_value:." + str(zeros) + "f}{suffix}").format(
        abs_value=abs_value, suffix=suffix
    )
    return f"-{formatted_value}" if value < 0 else f"+{formatted_value}"


gene_ticker = ticker.FuncFormatter(custom_gene_formatter)


def custom_formatter(value, tick_pos, base=1):
    if base != 1:
        value = value / base

    abs_value = abs(value)
    if abs_value >= 1000000:
        abs_value = abs_value / 1000000
        suffix = "mb"
    elif abs_value >= 1000:
        abs_value = abs_value / 1000
        suffix = "kb"
    elif abs_value == 0:
        return "0"
    else:
        suffix = "b"

    zeros = count_zeros(abs_value)

    formatted_value = ("{abs_value:." + str(zeros) + "f}{suffix}").format(
        abs_value=abs_value, suffix=suffix
    )
    return formatted_value


distance_ticker = ticker.FuncFormatter(custom_formatter)

import functools


def DistanceFormatter(base=1):
    return ticker.FuncFormatter(functools.partial(custom_formatter, base=base))


def lighten_color(color, amount=0.5):
    """
    Lightens the given color by multiplying (1-luminosity) by the given amount.
    Input can be matplotlib color string, hex string, or RGB tuple.

    Examples:
    >> lighten_color('g', 0.3)
    >> lighten_color('#F034A3', 0.6)
    >> lighten_color((.3,.55,.1), 0.5)
    """
    import matplotlib.colors as mc
    import colorsys

    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], 1 - amount * (1 - c[1]), c[2])



import numpy as np
import itertools
import dataclasses


@dataclasses.dataclass
class Minibatch:
    cells_oi: np.ndarray
    genes_oi: np.ndarray
    phase: str = "train"

    def items(self):
        return {"cells_oi": self.cells_oi, "genes_oi": self.genes_oi}

    def filter_genes(self, genes):
        genes_oi = self.genes_oi[genes[self.genes_oi]]

        return Minibatch(
            cells_oi=self.cells_oi,
            genes_oi=genes_oi,
            phase=self.phase,
        )


def create_bins_ordered(
    cells,
    genes,
    n_genes_total,
    n_genes_step=300,
    n_cells_step=1000,
    use_all=False,
    rg=None,
    permute_genes=True,
    permute_cells=True,
    **kwargs
):
    """
    Creates bins of cellxgene
    A number of cell and gene bins are created first, and all combinations of these bins make up the
    """
    if rg is None:
        rg = np.random.RandomState()
    if permute_cells:
        cells = rg.permutation(cells)
    cells = np.array(cells)
    if permute_genes:
        genes = rg.permutation(genes)
    genes = np.array(genes)

    gene_cuts = [*np.arange(0, len(genes), step=n_genes_step)]
    if use_all:
        gene_cuts.append(len(genes))
    gene_bins = [genes[a:b] for a, b in zip(gene_cuts[:-1], gene_cuts[1:])]

    cell_cuts = [*np.arange(0, len(cells), step=n_cells_step)]
    if use_all:
        cell_cuts.append(len(cells))
    cell_bins = [cells[a:b] for a, b in zip(cell_cuts[:-1], cell_cuts[1:])]

    bins = []
    for cells_oi, genes_oi in itertools.product(cell_bins, gene_bins):
        bins.append(
            Minibatch(
                cells_oi=cells_oi,
                genes_oi=genes_oi,
                # cellxgene_oi=cell_gene_to_cellxgene(cells_oi, genes_oi, n_genes_total),
                **kwargs
            )
        )
    return bins


def create_bins_random(
    cells,
    genes,
    n_genes_total,
    n_genes_step=300,
    n_cells_step=1000,
    use_all=False,
    rg=None,
    permute_genes=True,
    **kwargs
):
    """
    Creates bins of cellxgene
    Within each cell bin, the genes are put in a new set of random bins
    """
    if rg is None:
        rg = np.random.RandomState()
    cells = rg.permutation(cells)

    cell_cuts = [*np.arange(0, len(cells), step=n_cells_step)]
    if use_all:
        cell_cuts.append(len(cells))
    cell_bins = [cells[a:b] for a, b in zip(cell_cuts[:-1], cell_cuts[1:])]

    bins = []
    for cells_oi in cell_bins:
        if permute_genes:
            genes = rg.permutation(genes)
        gene_cuts = [*np.arange(0, len(genes), step=n_genes_step)]
        if use_all:
            gene_cuts.append(len(genes))
        gene_bins = [genes[a:b] for a, b in zip(gene_cuts[:-1], gene_cuts[1:])]

        for genes_oi in gene_bins:
            bins.append(
                Minibatch(
                    cells_oi=cells_oi,
                    genes_oi=genes_oi,
                    # cellxgene_oi=cell_gene_to_cellxgene(
                    #     cells_oi, genes_oi, n_genes_total
                    # ),
                    **kwargs
                )
            )
    return bins


class Minibatcher:
    def __init__(self, cells, genes, n_total_genes, n_cells_step, n_genes_step):
        self.cells = cells
        self.genes = genes
        self.n_total_genes = n_total_genes
        self.n_cells_step = n_cells_step
        self.n_genes_step = n_genes_step

    def create_minibatches(self, *args, **kwargs):
        minibatches = create_bins_ordered(
            self.cells,
            self.genes,
            self.n_total_genes,
            n_genes_step=self.n_genes_step,
            n_cells_step=self.n_cells_step,
            *args,
            **kwargs
        )
        return minibatches



# %% [markdown]
# ChromatinHD analyzes single-cell ATAC+RNA data using the raw fragments as input, by automatically adapting the scale at which
# relevant chromatin changes on a per-position, per-cell, and per-gene basis.
#
# Currently, the following models are supported:
#
# <ul>
#     <li><strong><i>pred</i></strong>: Predicting gene expression from fragments
#         <ul>
#             <li>To learn where accessibility is predictive for gene
#                 expression</li>
#             <li>To learn which regions in the genome are likely collaborating to regulate gene expression</li>
#             <li>To learn for which regions high fragment sizes are indicative of gene expression, which indicates an active regulatory region with dense protein binding</li>
#         </ul>
#     </li>
#     <li><strong><i>diff</i></strong>: Understanding the differences in accessibilty between cell types/states</li>
# </ul>



import copy
import threading
import numpy as np
import time


class ThreadWithResult(threading.Thread):
    result = None

    def __init__(
        self,
        group=None,
        target=None,
        name=None,
        loader=None,
        args=(),
        kwargs={},
        *,
        daemon=None
    ):
        assert target is not None
        self.loader = loader

        def function():
            self.result = target(*args, **kwargs)

        super().__init__(group=group, target=function, name=name, daemon=daemon)


class LoaderPool:
    tasks: list
    next_task_sets: list
    loaders_available: list
    n_todo: list[int]
    shuffle_on_iter = False

    def __init__(
        self,
        loader_cls,
        loader_kwargs=None,
        n_workers=3,
        shuffle_on_iter=False,
        loader=None,
    ):
        self.running = []

        self.shuffle_on_iter = shuffle_on_iter

        if loader_kwargs is None:
            loader_kwargs = {}
        self.loader_cls = loader_cls
        self.loader_kwargs = loader_kwargs

        self.n_workers = n_workers

        if loader is not None:
            self.loaders = [loader.copy() for i in range(n_workers)]
        else:
            self.loaders = [loader_cls(**loader_kwargs) for i in range(n_workers)]
        for loader in self.loaders:
            loader.running = False

    def initialize(self, tasks=None, next_task_sets=None, *args, **kwargs):
        if next_task_sets is None:
            next_task_sets = [{"tasks": tasks}]
        self.next_task_sets = copy.copy(next_task_sets)

        self.args = args
        self.kwargs = kwargs

        self.n_todo = []

        self.restart(*args, **kwargs)

    def restart_loaders(self, loader_kwargs=None):
        loader_cls = self.loader_cls
        if loader_kwargs is None:
            loader_kwargs = {}
        loader_kwargs = {**self.loader_kwargs, **loader_kwargs}
        self.loaders = [loader_cls(**loader_kwargs) for i in range(self.n_workers)]

        for loader in self.loaders:
            loader.running = False

        self.loaders_available = copy.copy(self.loaders)

    def submit(self, *args, **kwargs):
        if self.loaders_available is None:
            raise ValueError("Pool was not initialized")
        if len(self.loaders_available) == 0:
            raise ValueError("No loaders available")

        loader = self.loaders_available.pop(0)
        if loader.running:
            raise ValueError
        self.loaders_available = self.loaders_available + [loader]
        loader.running = True
        thread = ThreadWithResult(
            target=loader.load, loader=loader, args=args, kwargs=kwargs
        )
        self.running.append(thread)
        thread.start()

    def pull(self):
        thread = self.running.pop(0)
        # import time
        # start = time.time()
        thread.join()
        # end = time.time() - start
        # print(end)
        thread.loader.running = False
        self.n_done += 1
        return thread.result

    def __len__(self):
        return self.tasks.__len__()

    def __iter__(self):
        assert len(self.n_todo) > 0
        self.n_done = 0
        return self

    def __next__(self):
        if len(self.n_todo) == 0:
            raise StopIteration
        if self.n_done == self.n_todo[0]:
            self.n_todo.pop(0)
            raise StopIteration
        if len(self.running) == 0:
            if len(self.todo) > 0:
                raise ValueError(
                    "Iteration stopped too early, make sure to call submit_next in each iteration"
                )
        return self.pull()

    def submit_next(self):
        if len(self.todo) == 0:
            self.start_next()

        task = self.todo.pop(0)
        self.submit(task, *self.args, **self.kwargs)

    def start_next(self):
        if len(self.next_task_sets) > 0:
            # if some new tasks are next, recreate the tasks
            next_task_set = self.next_task_sets.pop(0)
            self.tasks = copy.copy(next_task_set["tasks"])
            if ("loader_kwargs" in next_task_set) and (
                next_task_set["loader_kwargs"] is not None
            ):
                self.restart_loaders(loader_kwargs=next_task_set["loader_kwargs"])

        self.n_todo.append(len(self.tasks))

        if self.shuffle_on_iter:
            self.todo = [
                self.tasks[i]
                for i in self.rg.choice(len(self.tasks), len(self.tasks), replace=False)
            ]
        else:
            self.todo = copy.copy(self.tasks)

    def restart(self):
        # join all still running threads
        for thread in self.running:
            thread.join()

        if self.shuffle_on_iter:
            self.rg = np.random.RandomState(0)

        for loader in self.loaders:
            loader.running = False

        self.loaders_available = copy.copy(self.loaders)

        self.n_done = 0
        self.running = []
        self.todo = []
        self.n_todo = []

        self.start_next()

        for i in range(min(len(self.tasks), self.n_workers - len(self.running) - 1)):
            self.submit_next()

    def terminate(self):
        for thread in self.running:
            thread.join()
        self.running = []
        self.loaders = None

    def __del__(self):
        self.terminate()


class LoaderPool2:
    loaders_available: list
    n_todo: list[int]
    shuffle_on_iter = False

    def __init__(
        self,
        loader_cls,
        loader_kwargs=None,
        n_workers=3,
        loader=None,
    ):
        self.running = []

        if loader_kwargs is None:
            loader_kwargs = {}
        self.loader_cls = loader_cls
        self.loader_kwargs = loader_kwargs

        self.n_workers = n_workers

        if loader is not None:
            self.loaders = [loader.copy() for i in range(n_workers)]
        else:
            self.loaders = [loader_cls(**loader_kwargs) for i in range(n_workers)]
        for loader in self.loaders:
            loader.running = False
        self.wait = []

    def initialize(self, tasker, *args, **kwargs):
        self.tasker = tasker

        self.args = args
        self.kwargs = kwargs

        self.restart(*args, **kwargs)

    def submit_next(self):
        try:
            task = next(self.iter)
        except StopIteration:
            return
        self.submit(task, *self.args, **self.kwargs)

    def submit(self, *args, **kwargs):
        if self.loaders_available is None:
            raise ValueError("Pool was not initialized")
        if len(self.loaders_available) == 0:
            raise ValueError("No loaders available")

        loader = self.loaders_available.pop(0)
        if loader.running:
            raise ValueError
        self.loaders_available = self.loaders_available + [loader]
        loader.running = True
        thread = ThreadWithResult(
            target=loader.load, loader=loader, args=args, kwargs=kwargs
        )
        self.running.append(thread)
        thread.start()

    def pull(self):
        thread = self.running.pop(0)
        import time

        start = time.time()
        thread.join()
        self.wait.append(time.time() - start)
        thread.loader.running = False
        self.submit_next()
        return thread.result

    def __len__(self):
        return self.tasker.__len__()

    def __iter__(self):
        self.restart()
        return self

    def __next__(self):
        if len(self.running) == 0:
            # if len(self.wait) > 0:
            #     print("Average wait: ", np.mean(self.wait))
            #     print("Cumulative wait: ", np.sum(self.wait))
            raise StopIteration
        return self.pull()

    def restart(self):
        # join all still running threads
        for thread in self.running:
            thread.join()

        for loader in self.loaders:
            loader.running = False

        self.loaders_available = copy.copy(self.loaders)

        self.running = []

        self.wait = []

        self.iter = iter(self.tasker)

        for i in range(min(len(self.tasker), self.n_workers - len(self.running) - 1)):
            self.submit_next()

    def terminate(self):
        for thread in self.running:
            thread.join()
        self.running = []
        self.loaders = None

    def __del__(self):
        self.terminate()



import torch
import tqdm.auto as tqdm
import math
from chromatinhd.models.diff.model import splines
from chromatinhd.embedding import EmbeddingTensor


class TransformedDistribution(torch.nn.Module):
    def __init__(self, transform):
        super().__init__()
        self.transform = transform

    def log_prob(self, x, *args, **kwargs):
        log_prob = torch.zeros_like(x)
        x_ = x
        x_, logabsdet = self.transform.transform_forward(x_, *args, **kwargs)
        log_prob = log_prob + logabsdet
        return log_prob

    def sample(self, sample_shape=torch.Size(), *args, device=None, **kwargs):
        y = torch.rand(sample_shape, device=device)
        y, _ = self.transform.transform_inverse(y, *args, **kwargs)
        return y

    def parameters_sparse(self):
        return self.transform.parameters_sparse()

    def parameters_dense(self):
        return self.transform.parameters_dense()


class QuadraticSplineTransform(torch.nn.Module):
    bijective = True
    domain = torch.distributions.constraints.real
    codomain = torch.distributions.constraints.real

    def __init__(self, unnormalized_widths, unnormalized_heights):
        super().__init__()
        self.unnormalized_widths = torch.nn.Parameter(unnormalized_widths)
        self.unnormalized_heights = torch.nn.Parameter(unnormalized_heights)

    def transform_forward(self, x, local_gene_ix, inverse=False):
        widths = splines.quadratic.calculate_widths(self.unnormalized_widths)
        heights = splines.quadratic.calculate_heights(self.unnormalized_heights, widths)
        bin_left_cdf = splines.quadratic.calculate_bin_left_cdf(heights, widths)
        bin_locations = splines.quadratic.calculate_bin_locations(widths)

        outputs, logabsdet = splines.quadratic.quadratic_spline(
            x,
            widths=widths[local_gene_ix],
            heights=heights[local_gene_ix],
            bin_left_cdf=bin_left_cdf[local_gene_ix],
            bin_locations=bin_locations[local_gene_ix],
            inverse=inverse,
        )
        return outputs, logabsdet

    def transform_inverse(self, y, local_gene_ix):
        return self.transform_inverse(y, local_gene_ix, inverse=True)


def prioritize(x, n, k=2):
    possible = torch.linspace(0, 1, n * k + 2)[1:-1]

    y_ = torch.linspace(0, 1, n * k)
    dist = torch.distributions.Normal(0.0, scale=(1 / (n * k * 2)))

    weights = dist.log_prob(torch.sqrt((x - y_.unsqueeze(1)) ** 2))
    density = torch.logsumexp(weights, 1)

    diff1 = torch.nn.functional.pad(torch.exp(density).diff(), (1, 0))
    diff1 = torch.nn.functional.pad(torch.exp(density).diff().diff(), (1, 1))
    possible_scores = diff1.reshape((n * k, 1)).mean(1)

    chosen_idx = torch.argsort(-possible_scores.abs())[: (n - 2)]
    chosen = torch.nn.functional.pad(torch.sort(possible[chosen_idx])[0], (1, 1))
    chosen[..., -1] = 1.0
    return chosen


def initialize_from_previous(x, n, local_gene_ix, n_genes, transforms, device="cuda"):
    q2_orig = torch.linspace(0, 1, n).expand(n_genes, -1)
    # q2_orig = prioritize(x, n).expand(n_genes, -1)
    q2 = q2_orig

    # calculate bin widths
    for transform in transforms:
        q2 = (
            transform.transform_forward(
                q2.flatten(),
                local_gene_ix=torch.repeat_interleave(
                    torch.ones(n_genes, dtype=int) * n
                ),
            )[0]
            .detach()
            .reshape(q2.shape)
        )
        # transforms may lead to equivalent consecutive bins
        # which leads to NaN log(width)
        # therefore add some eps to these bins
        if (q2.diff() <= 0).any():
            eps = 1e-6
            q2 = torch.nn.functional.pad(torch.cumsum((q2.diff() + eps), 1), (1, 0))
            q2 = q2 / q2.diff().sum(1, keepdim=True)
            assert (q2.diff() > 0).all()

    unnormalized_widths = torch.log(q2.diff())

    # calculate the bincount in chunks
    # runs on a device
    chunk_width = int(1e6)
    bincount = torch.zeros((n_genes, n - 1), dtype=int)

    transforms = [transform.to(device) for transform in transforms]

    q2 = q2.to(device)

    for x2, local_gene_ix2 in tqdm.tqdm(
        zip(x.split(chunk_width, 0), local_gene_ix.split(chunk_width, 0)),
        total=math.ceil(x.shape[0] / chunk_width),
    ):
        x2 = x2.to(device)
        local_gene_ix2 = local_gene_ix2.to(device)
        for transform in transforms:
            x2 = transform.transform_forward(x2, local_gene_ix2)[0].detach()

        digitized = torch.clamp(
            torch.searchsorted(
                q2[local_gene_ix2], x2.unsqueeze(-1), right=True
            ).squeeze(-1),
            0,
            n - 1,
        )
        bincount += (
            torch.bincount((digitized + local_gene_ix2 * n), minlength=n * n_genes)
            .reshape((n_genes, n))[:, 1:]
            .cpu()
        )

    q2 = q2.to("cpu")
    transforms = [transform.to("cpu") for transform in transforms]

    # calculate the initial bin height (=pdf) by taking the average bincount around each knot
    aroundcounts = torch.nn.functional.pad(
        bincount / q2.diff(), (1, 0)
    ) + torch.nn.functional.pad(bincount / q2.diff(), (0, 1))
    unnormalized_heights = torch.log(
        aroundcounts + 1e-2
    )  # small pseudocount for those bins without a single count

    if unnormalized_heights.isnan().any():
        raise ValueError("NaNs in initialized pdf")

    return unnormalized_heights, unnormalized_widths


class QuadraticSplineStack(torch.nn.Module):
    def __init__(self, x, nbins, local_gene_ix, n_genes):
        super().__init__()
        splits = []
        unnormalized = []
        transforms = []
        for n in nbins:
            unnormalized_heights, unnormalized_widths = initialize_from_previous(
                x, n, local_gene_ix, n_genes, transforms=transforms
            )
            unnormalized.extend([unnormalized_heights, unnormalized_widths])
            splits.extend(
                [unnormalized_heights.shape[-1], unnormalized_widths.shape[-1]]
            )
            transforms.append(
                QuadraticSplineTransform(unnormalized_widths, unnormalized_heights)
            )
        self.unnormalized = torch.nn.Parameter(torch.cat(unnormalized, -1))
        self.splits = splits

    def _split_parameters(self, x):
        split = x.split(self.splits, -1)
        return zip(split[0::2], split[1::2])

    def transform_forward(self, x, local_gene_ix, inverse=False):
        assert x.shape == local_gene_ix.shape

        logabsdet = None
        outputs = x
        for unnormalized_heights, unnormalized_widths in self._split_parameters(
            self.unnormalized
        ):
            widths = splines.quadratic.calculate_widths(unnormalized_widths)
            bin_locations = splines.quadratic.calculate_bin_locations(widths)

            heights = splines.quadratic.calculate_heights(unnormalized_heights, widths)
            bin_left_cdf = splines.quadratic.calculate_bin_left_cdf(heights, widths)

            outputs, logabsdet_ = splines.quadratic.quadratic_spline(
                outputs,
                widths=widths[local_gene_ix],
                heights=heights[local_gene_ix],
                bin_left_cdf=bin_left_cdf[local_gene_ix],
                bin_locations=bin_locations[local_gene_ix],
                inverse=inverse,
            )
            if logabsdet is None:
                logabsdet = logabsdet_
            else:
                logabsdet = logabsdet + logabsdet_
        return outputs, logabsdet

    def transform_inverse(self, y, local_gene_ix):
        return self.transform_forward(y, local_gene_ix=local_gene_ix, inverse=True)


class DifferentialQuadraticSplineStack(torch.nn.Module):
    def __init__(self, nbins, n_genes, local_gene_ix=None, x=None):
        self.nbins = nbins

        super().__init__()
        unnormalized_heights = []
        unnormalized_widths = []
        splits_heights = []
        splits_widths = []
        transforms = []
        split_deltas = []
        for n in nbins:
            n_heights = n
            n_widths = n - 1

            if x is not None:
                unnormalized_heights_, unnormalized_widths_ = initialize_from_previous(
                    x, n, local_gene_ix, n_genes, transforms=transforms
                )
                unnormalized_heights.append(unnormalized_heights_)
                unnormalized_widths.append(unnormalized_widths_)

                assert unnormalized_heights_.shape[-1] == n_heights
                assert unnormalized_widths_.shape[-1] == n_widths

                transforms.append(
                    QuadraticSplineTransform(
                        unnormalized_widths_, unnormalized_heights_
                    )
                )
            splits_heights.append(n_heights)
            splits_widths.append(n_widths)
            split_deltas.append(n_heights)

        self.unnormalized_heights = EmbeddingTensor(
            n_genes, (sum(splits_heights),), sparse=True
        )

        self.unnormalized_widths = EmbeddingTensor(
            n_genes, (sum(splits_widths),), sparse=True
        )

        if x is not None:
            unnormalized_heights = torch.cat(unnormalized_heights, -1)
            unnormalized_widths = torch.cat(unnormalized_widths, -1)
            self.unnormalized_heights.data = unnormalized_heights
            self.unnormalized_widths.data = unnormalized_widths
        else:
            self.unnormalized_heights.data[:] = 0.0
            self.unnormalized_widths.data[:] = 0.0

        self.splits_heights = splits_heights
        self.splits_widths = splits_widths
        self.split_deltas = split_deltas

    def _split_parameters(self, x, splits):
        return x.split(splits, -1)

    def transform_forward(self, x, genes_oi, local_gene_ix, delta, inverse=False):
        assert x.shape == local_gene_ix.shape

        logabsdet = None
        outputs = x

        stride = 1 if not inverse else -1

        unnormalized_widths = self.unnormalized_widths(genes_oi)
        unnormalized_heights = self.unnormalized_heights(genes_oi)
        for unnormalized_heights, unnormalized_widths, delta_heights in zip(
            self._split_parameters(unnormalized_heights, self.splits_heights)[::stride],
            self._split_parameters(unnormalized_widths, self.splits_widths)[::stride],
            self._split_parameters(delta, self.split_deltas)[::stride],
        ):
            widths = splines.quadratic.calculate_widths(unnormalized_widths)
            bin_locations = splines.quadratic.calculate_bin_locations(widths)

            # use index_select here as it is much faster in backwards than regular indexing
            widths = widths.index_select(0, local_gene_ix)
            bin_locations = bin_locations.index_select(0, local_gene_ix)

            unnormalized_heights = (
                unnormalized_heights.index_select(0, local_gene_ix) + delta_heights
            )
            heights = splines.quadratic.calculate_heights(unnormalized_heights, widths)
            bin_left_cdf = splines.quadratic.calculate_bin_left_cdf(heights, widths)

            outputs, logabsdet_ = splines.quadratic.quadratic_spline(
                outputs,
                widths=widths,
                heights=heights,
                bin_left_cdf=bin_left_cdf,
                bin_locations=bin_locations,
                inverse=inverse,
            )
            if logabsdet is None:
                logabsdet = logabsdet_
            else:
                logabsdet = logabsdet + logabsdet_
        return outputs, logabsdet

    def transform_inverse(self, y, genes_oi, local_gene_ix, delta):
        return self.transform_forward(
            y, genes_oi=genes_oi, local_gene_ix=local_gene_ix, delta=delta, inverse=True
        )

    def parameters_sparse(self):
        return [self.unnormalized_heights.weight, self.unnormalized_widths.weight]

    def parameters_dense(self):
        return []



import chromatinhd.grid
import matplotlib as mpl
from chromatinhd.grid.broken import Broken, Panel
import numpy as np


class Effect(chromatinhd.grid.Panel):
    def __init__(self, plotdata, window, width, show_accessibility=False):
        super().__init__((width, 0.5))

        ax = self.ax
        ax.set_xlim(*window)

        ax.plot(
            plotdata["position"],
            plotdata["effect"],
            color="#333",
            lw=1,
        )
        ax.fill_between(
            plotdata["position"],
            plotdata["effect"],
            0,
            color="#333",
            alpha=0.2,
            lw=0,
        )

        ax.set_ylabel(
            "Effect",
            rotation=0,
            ha="right",
            va="center",
        )

        ax.set_xticks([])
        ax.invert_yaxis()
        ymax = plotdata["effect"].abs().max()
        ax.set_ylim(-ymax, ymax)

        # change vertical alignment of last y tick to bottom
        ax.set_yticks([0, ax.get_ylim()[1]])
        ax.get_yticklabels()[-1].set_verticalalignment("top")
        ax.get_yticklabels()[0].set_verticalalignment("bottom")

        # vline at tss
        ax.axvline(0, color="#888888", lw=0.5, zorder=-1, dashes=(2, 2))

    @classmethod
    def from_genemultiwindow(
        cls, genemultiwindow, gene, width, show_accessibility=False
    ):
        plotdata = genemultiwindow.get_plotdata(gene).reset_index()
        window = np.array([plotdata["position"].min(), plotdata["position"].max()])
        return cls(plotdata, window, width, show_accessibility=show_accessibility)



from chromatinhd.flow import Flow, CompressedNumpyInt64, CompressedNumpyFloat64, Stored
import pandas as pd


class Motiftrack(Flow):
    scores = CompressedNumpyFloat64("scores")

    _motifs = None

    @property
    def motifs(self):
        if self._motifs is None:
            self._motifs = pd.read_pickle(self.path / "motifs.pkl")
        return self._motifs

    @motifs.setter
    def motifs(self, value):
        value.index.name = "gene"
        value.to_pickle(self.path / "motifs.pkl")
        self._motifs = value



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.5
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Preprocess

# %%
from IPython import get_ipython

if get_ipython():
    get_ipython().run_line_magic("load_ext", "autoreload")
    get_ipython().run_line_magic("autoreload", "2")

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib as mpl

import seaborn as sns

sns.set_style("ticks")

import torch

import pickle

import scanpy as sc

import tqdm.auto as tqdm
import io
import pathlib

# %%
import chromatinhd as chd

# %%
# install the genome if it doesn't exist yet
# this can take a couple of minutes
import genomepy
genomepy.install_genome("GRCh38")

# %%
folder_data = chd.get_git_root() / "data"
raw_folder = chd.get_git_root() / "tmp" / "pbmc10k"

dataset_name = "tiny"
main_url = "https://cf.10xgenomics.com/samples/cell-arc/2.0.0/pbmc_granulocyte_sorted_10k/pbmc_granulocyte_sorted_10k"
genome_name = "GRCh38"
organism = "hs"

folder_data_preproc = raw_folder
folder_data_preproc.mkdir(exist_ok=True, parents=True)

genome = genomepy.Genome("GRCh38")

# %% [markdown]
# ## Download

# %%
if not (folder_data_preproc / "filtered_feature_bc_matrix.h5").exists():
    ! wget {main_url}_filtered_feature_bc_matrix.h5 -O {folder_data_preproc}/filtered_feature_bc_matrix.h5

# %%
if not (folder_data_preproc / "atac_fragments.tsv.gz").exists():
    !wget {main_url}_atac_fragments.tsv.gz -O {folder_data_preproc}/atac_fragments.tsv.gz

# %%
if not (folder_data_preproc / "atac_fragments.tsv.gz.tbi").exists():
    !wget {main_url}_atac_fragments.tsv.gz.tbi -O {folder_data_preproc}/atac_fragments.tsv.gz.tbi

# %%
!ls -lh {folder_data_preproc}

# %% [markdown]
# ### Genes

# %% [markdown]
# Get the gene annotaton from ensembl biomart

# %%
biomart_dataset = chd.biomart.Dataset()
genes = biomart_dataset.get([
    biomart_dataset.attribute("ensembl_gene_id"),
    biomart_dataset.attribute("transcript_start"),
    biomart_dataset.attribute("transcript_end"),
    biomart_dataset.attribute("end_position"),
    biomart_dataset.attribute("start_position"),
    biomart_dataset.attribute("ensembl_transcript_id"),
    biomart_dataset.attribute("chromosome_name"),
    biomart_dataset.attribute("strand"),
    biomart_dataset.attribute("external_gene_name"),
    biomart_dataset.attribute("transcript_is_canonical"),
    biomart_dataset.attribute("transcript_biotype")
], filters = [
    biomart_dataset.filter("transcript_biotype", value = "protein_coding"),
])
genes["chrom"] = "chr" + genes["chromosome_name"].astype(str)
genes = genes.sort_values("transcript_is_canonical").groupby("ensembl_gene_id").first()
genes = genes.rename(
    columns = {
        "transcript_start": "start",
        "transcript_end": "end",
        "external_gene_name": "symbol",
    }
)
genes.index.name = "gene"
genes = genes.drop(["chromosome_name", "transcript_biotype", "transcript_is_canonical", "start_position", "end_position"], axis = 1)

# remove genes without a symbol
genes = genes.loc[~pd.isnull(genes["symbol"])]

# remove genes on alternative assemblies
genes = genes.loc[~genes["chrom"].str.contains("_")]

# %%
assert (
    genes.groupby(level=0).size().mean() == 1
), "For each gene, there should only be one transcript"

# %%
genes.to_csv(folder_data_preproc / "genes.csv")

# %% [markdown]
# ## Create transcriptome

# %% [markdown]
# Read in the transcriptome

# %%
genes = pd.read_csv(folder_data_preproc / "genes.csv", index_col=0)
transcriptome = chd.data.Transcriptome(folder_data_preproc / "transcriptome")

# %% [markdown]
# ### Read and process

# %%
adata = sc.read_10x_h5(folder_data_preproc / "filtered_feature_bc_matrix.h5")
adata.var.index.name = "symbol"
adata.var = adata.var.reset_index()
adata.var.index = adata.var["gene_ids"]
adata.var.index.name = "gene"

all_gene_ids = sorted(
    list(set(genes.loc[genes["chrom"] != "chrMT"].index) & set(adata.var.index))
)

adata = adata[:, all_gene_ids]

# %%
adata.var_names_make_unique()

# %%
print(adata.obs.shape[0])
sc.pp.filter_cells(adata, min_counts=1000)
print(adata.obs.shape[0])
sc.pp.filter_cells(adata, min_genes=200)
print(adata.obs.shape[0])
sc.pp.filter_genes(adata, min_cells=3)
print(adata.var.shape[0])

# %%
# you may have to install scrublet, `pip install scrublet`
sc.external.pp.scrublet(adata)

# %%
adata.obs["doublet_score"].plot(kind="hist")

# %%
adata.obs["doublet"] = (adata.obs["doublet_score"] > 0.1).astype("category")

print(adata.obs.shape[0])
adata = adata[~adata.obs["doublet"].astype(bool)]
print(adata.obs.shape[0])

# %%
size_factor = np.median(np.array(adata.X.sum(1)))
adata.uns["size_factor"] = size_factor

# %%
sc.pp.normalize_total(adata, size_factor)

# %%
sc.pp.log1p(adata)

# %%
sc.pp.pca(adata)

# %%
sc.pp.highly_variable_genes(adata)

# %%
adata.var["n_cells"] = np.array((adata.X > 0).sum(0))[0]

# %%
# adata = adata[:, adata.var["dispersions_norm"].sort_values(ascending = False)[:5000].index]
print(adata.var.shape[0])

genes_oi = (
    adata.var.query("n_cells > 100")["dispersions_norm"]
    .sort_values(ascending=False)[:5000]
    .index.tolist()
)

adata = adata[:, genes_oi]
print(adata.var.shape[0])

all_gene_ids = adata.var.index

# %%
sc.pp.neighbors(adata)
sc.tl.umap(adata)

# %%
adata.var["chrom"] = genes["chrom"]

# %%
# may have to install leiden algorithm, `pip install leidenalg`
sc.tl.leiden(adata, resolution=1.0)

# %%
sc.pl.umap(adata, color="leiden", legend_loc="on data")

# %%
adata = adata[~adata.obs["leiden"].isin(["15", "16"])]

# %%
sc.pl.umap(adata, color="leiden")

# %%
transcriptome.adata = adata
transcriptome.var = adata.var
transcriptome.obs = adata.obs

# %%
transcriptome.layers["X"] = adata.X

# %%
fig, ax = plt.subplots()
sns.scatterplot(adata.var, x="means", y="dispersions_norm")
ax.set_xscale("log")
# %%
# may have to install magic, `pip install magic-impute`
import magic

magic_operator = magic.MAGIC(knn=30)
X_smoothened = magic_operator.fit_transform(adata.X)
adata.layers["magic"] = X_smoothened

# %%
transcriptome.layers["magic"] = X_smoothened

# %%
genes_oi = adata.var.sort_values("dispersions_norm", ascending=False).index[:10]
genes_oi = transcriptome.gene_id(["CCL4"])
sc.pl.umap(adata, color=genes_oi, title=transcriptome.symbol(genes_oi), layer="magic")

# %%
sc.pl.umap(adata, color="leiden", legend_loc="on data")

# %% [markdown]
# ### Interpret PBMC10K

# %%
sc.tl.rank_genes_groups(adata, "leiden", method="t-test")
# sc.pl.rank_genes_groups(adata, n_genes=25, sharey=False, gene_symbols = "symbol")

# %%
import io

# %%
marker_annotation = pd.read_table(
    io.StringIO(
        """ix	symbols	celltype
0	IL7R, CD3D	CD4 naive T
0	IL7R, CD3D, ITGB1	CD4 memory T
1	CD14, LYZ	CD14+ Monocytes
2	MS4A1, IL4R, CD79A	naive B
2	MS4A1, CD79A, TNFRSF13B	memory B
3	CD8A, CD3D	CD8 naive T
4	GNLY, NKG7, GZMA, GZMB, NCAM1	NK
4	GNLY, NKG7, CD3D, CCL5, GZMA, CD8A	CD8 activated T
4	SLC4A10	MAIT
5	FCGR3A, MS4A7	FCGR3A+ Monocytes
5	CD27, JCHAIN	Plasma
6	TCF4	pDCs
6	CST3, CD1C	cDCs
"""
    )
).set_index("celltype")
marker_annotation["symbols"] = marker_annotation["symbols"].str.split(", ")
# marker_annotation = marker_annotation.explode("symbols")

# %%
sc.pl.umap(
    adata,
    color=transcriptome.gene_id(
        marker_annotation.query("celltype == 'cDCs'")["symbols"].explode()
    ),
)
# %%
sc.pl.umap(
    adata,
    color=transcriptome.gene_id(
        "CD1C"
    ),
)


# %%
# Define cluster score for all markers
def evaluate_partition(
    anndata, marker_dict, gene_symbol_key=None, partition_key="louvain_r1"
):
    # Inputs:
    #    anndata         - An AnnData object containing the data set and a partition
    #    marker_dict     - A dictionary with cell-type markers. The markers should be stores as anndata.var_names or
    #                      an anndata.var field with the key given by the gene_symbol_key input
    #    gene_symbol_key - The key for the anndata.var field with gene IDs or names that correspond to the marker
    #                      genes
    #    partition_key   - The key for the anndata.obs field where the cluster IDs are stored. The default is
    #                      'louvain_r1'

    # Test inputs
    if partition_key not in anndata.obs.columns.values:
        print("KeyError: The partition key was not found in the passed AnnData object.")
        print(
            "   Have you done the clustering? If so, please tell pass the cluster IDs with the AnnData object!"
        )
        raise

    if (gene_symbol_key != None) and (
        gene_symbol_key not in anndata.var.columns.values
    ):
        print(
            "KeyError: The provided gene symbol key was not found in the passed AnnData object."
        )
        print(
            "   Check that your cell type markers are given in a format that your anndata object knows!"
        )
        raise

    if gene_symbol_key:
        gene_ids = anndata.var[gene_symbol_key]
    else:
        gene_ids = anndata.var_names

    clusters = np.unique(anndata.obs[partition_key])
    n_clust = len(clusters)
    n_groups = len(marker_dict)

    marker_res = np.zeros((n_groups, n_clust))
    z_scores = sc.pp.scale(anndata, copy=True)

    i = 0
    for group in marker_dict:
        # Find the corresponding columns and get their mean expression in the cluster
        j = 0
        for clust in clusters:
            cluster_cells = np.in1d(z_scores.obs[partition_key], clust)
            marker_genes = np.in1d(gene_ids, marker_dict[group])
            marker_res[i, j] = z_scores.X[np.ix_(cluster_cells, marker_genes)].mean()
            j += 1
        i += 1

    variances = np.nanvar(marker_res, axis=0)
    if np.all(np.isnan(variances)):
        print(
            "No variances could be computed, check if your cell markers are in the data set."
        )
        print(
            "Maybe the cell marker IDs do not correspond to your gene_symbol_key input or the var_names"
        )
        raise

    marker_res_df = pd.DataFrame(marker_res, columns=clusters, index=marker_dict.keys())

    # Return the median of the variances over the clusters
    return marker_res_df


# %%
cluster_celltypes = evaluate_partition(
    adata, marker_annotation["symbols"].to_dict(), "symbol", partition_key="leiden"
).idxmax()

# %%
transcriptome.adata.obs["celltype"] = adata.obs["celltype"] = cluster_celltypes[
    adata.obs["leiden"]
].values
transcriptome.adata.obs["celltype"] = adata.obs["celltype"] = adata.obs[
    "celltype"
].astype(str)
# adata.obs.loc[adata.obs["leiden"] == "4", "celltype"] = "NKT"

# %%
transcriptome.adata.obs["log_n_counts"] = np.log(transcriptome.adata.obs["n_counts"])

# %%
sc.pl.umap(transcriptome.adata, color=["celltype", "log_n_counts", "leiden"])
sc.pl.umap(
    transcriptome.adata,
    color=transcriptome.gene_id(marker_annotation["symbols"].explode()),
    title=marker_annotation["symbols"].explode(),
)
# %%
transcriptome.obs = adata.obs
transcriptome.adata = adata

# %% [markdown]
# ## Create windows

# %% [markdown]
# ### Read genes and transcriptome

# %%
genes = pd.read_csv(folder_data_preproc / "genes.csv", index_col=0)
transcriptome = chd.data.Transcriptome(folder_data_preproc / "transcriptome")

# %% [markdown]
# ### Creating regions

# %%
region_name, region_window = "10k10k", (-10000, 10000)

# %%
all_gene_ids = transcriptome.var.index

# %%
regions = chd.data.Regions.from_genes(
    genes.loc[transcriptome.var.index],
    region_window,
    path = folder_data_preproc / "regions" / region_name
)

# %% [markdown]
# #### Create fragments
fragments = chd.data.Fragments.from_fragments_tsv(
    folder_data_preproc / "atac_fragments.tsv.gz",
    regions,
    transcriptome.obs,
    path = folder_data_preproc / "fragments" / regions.path.name,
)

# %%
fragments = chd.data.Fragments(folder_data_preproc / "fragments" / regions.path.name)

# %%
fragments.create_cellxgene_indptr()

# %%
fig, ax = plt.subplots()
bins = np.arange(0, 1000, 10)
sizes = fragments.coordinates[:, 1] - fragments.coordinates[:, 0]
bincount = np.bincount(np.digitize(sizes, bins))
ax.bar(bins, bincount[:-1], width=10, lw = 0)

# %% [markdown]
# ## Add MAGIC

# %%
transcriptome = chd.data.Transcriptome(folder_data_preproc / "transcriptome")

# %%
adata = transcriptome.adata

# %%
import magic

magic_operator = magic.MAGIC(knn=30)
X_smoothened = magic_operator.fit_transform(adata.X)

# %%
transcriptome.layers["magic"] = X_smoothened

# %% [markdown]
# ## Store small dataset

# %%
folder_dataset = chd.get_git_root() / "src" / "chromatinhd" / "data" / "examples" / "pbmc10ktiny"
folder_dataset.mkdir(exist_ok=True, parents=True)

# %%
transcriptome_original = chd.data.Transcriptome(folder_data_preproc / "transcriptome")
fragments_original = chd.data.Fragments(folder_data_preproc / "fragments" / "10k10k")

# %%
genes_oi = transcriptome_original.var.sort_values("dispersions_norm", ascending=False).index[:50]
# genes_oi = transcriptome_original.var.loc[transcriptome_original.gene_id(["IL1B"])].index
adata = sc.AnnData(
    X = transcriptome_original.adata[:, genes_oi].layers["magic"],
    obs = transcriptome_original.obs,
    var = transcriptome_original.var.loc[genes_oi]
)


# %%
adata.write(folder_dataset / "transcriptome.h5ad", compression="gzip")

# %%
import pysam
fragments_tabix = pysam.TabixFile(str(folder_data_preproc / "atac_fragments.tsv.gz"))

coordinates = fragments_original.regions.coordinates.loc[adata.var.index]
# %%
fragments_new = []
for i, (gene, promoter_info) in tqdm.tqdm(
    enumerate(coordinates.iterrows()), total=coordinates.shape[0]
):
    start = max(0, promoter_info["start"])
    
    fragments_promoter = fragments_tabix.fetch(promoter_info["chrom"], start, promoter_info["end"])
    fragments_new.extend(list(fragments_promoter))

fragments_new = pd.DataFrame([x.split("\t") for x in fragments_new], columns = ["chrom", "start", "end", "cell", "nreads"])
fragments_new["start"] = fragments_new["start"].astype(int)
fragments_new = fragments_new.sort_values(["chrom", "start", "cell"])

# %%
import gzip
fragments_new.to_csv(folder_dataset / "fragments.tsv", sep="\t", index=False, header = False)

# %%
import pysam
pysam.tabix_compress(folder_dataset / "fragments.tsv", folder_dataset / "fragments.tsv.gz", force=True)
# %%
!ls -lh {folder_dataset}
# %%
!tabix -p bed {folder_dataset / "fragments.tsv.gz"}
# %%
folder_dataset
# %%


# %% [markdown]
# ## Try to create dataset

# %%
gene_ix = transcriptome_original.gene_ix("IL1B")
transcriptome_original.obs["n_fragments"] = torch.bincount(fragments_original.mapping[fragments_original.mapping[:, 1] == gene_ix, 0], minlength = transcriptome_original.obs.shape[0])
transcriptome_original.obs["expression"] = transcriptome_original.layers["magic"][:, gene_ix]

# %%
fig, ax = plt.subplots()
sns.scatterplot(transcriptome_original.obs, x="n_fragments", y="expression", alpha = 0.1, s = 1, ax = ax)
# %%
import pathlib
import shutil

dataset_folder = pathlib.Path("/tmp/chromatinhd/pbmc10ktiny")
dataset_folder.mkdir(exist_ok=True, parents=True)

for file in dataset_folder.iterdir():
    if file.is_file():
        file.unlink()
    else:
        shutil.rmtree(file)
# %%
import pkg_resources

DATA_PATH = pathlib.Path(
    pkg_resources.resource_filename("chromatinhd", "data/examples/pbmc10ktiny/")
)

# copy all files from data path to dataset folder
for file in DATA_PATH.iterdir():
    shutil.copy(file, dataset_folder / file.name)
# %%
!ls {dataset_folder}
# %%
import scanpy as sc

adata = sc.read(dataset_folder / "transcriptome.h5ad")
# %%
transcriptome = chd.data.Transcriptome.from_adata(
    adata, path=dataset_folder / "transcriptome"
)

# %%
biomart_dataset = chd.biomart.Dataset.from_genome("GRCh38")
canonical_transcripts = chd.biomart.get_canonical_transcripts(
    biomart_dataset, transcriptome.var.index
)
# %%
regions = chd.data.Regions.from_canonical_transcripts(
    canonical_transcripts,
    path=dataset_folder / "regions",
    window=[-10000, 10000],
)
# %%
if not (dataset_folder / "fragments.tsv.gz.tbi").exists():
    !tabix {dataset_folder}/fragments.tsv.gz
# %%
fragments = chd.data.Fragments.from_fragments_tsv(
    dataset_folder / "fragments.tsv.gz",
    regions,
    obs = transcriptome.obs,
    path=dataset_folder / "fragments",
)
# %%
fragments.mapping.shape

# %%
gene_ix = transcriptome.gene_ix("IL1B")
transcriptome.obs["n_fragments"] = torch.bincount(fragments.mapping[fragments.mapping[:, 1] == gene_ix, 0], minlength = transcriptome.obs.shape[0])
transcriptome.obs["expression"] = transcriptome.layers["X"][:, gene_ix]

# %%
fig, ax = plt.subplots()
sns.regplot(transcriptome.obs, x="n_fragments", y="expression", ax = ax)
# %%



class HybridModel:
    def parameters_dense(self, autoextend=True):
        """
        Get all dense parameters of the model
        """
        parameters = [
            parameter
            for module in self._modules.values()
            for parameter in (
                module.parameters_dense() if hasattr(module, "parameters_dense") else []
            )
        ]

        # extend with any left over parameters that were not specified in parameters_dense or parameters_sparse
        def contains(x, y):
            return any([x is y_ for y_ in y])

        if autoextend:
            parameters.extend(
                [
                    p
                    for p in self.parameters()
                    if (not contains(p, self.parameters_sparse()))
                    and (not contains(p, parameters))
                ]
            )
        return parameters

    def parameters_sparse(self):
        """
        Get all sparse parameters in a model
        """
        return [
            parameter
            for module in self._modules.values()
            for parameter in (
                module.parameters_sparse()
                if hasattr(module, "parameters_sparse")
                else []
            )
        ]

from . import pred
from . import diff


import numpy as np
import dataclasses
import itertools
import math
import torch


@dataclasses.dataclass
class Minibatch:
    cells_oi: np.ndarray
    genes_oi: np.ndarray
    phase: str = "train"
    device: str = "cpu"

    def items(self):
        return {"cells_oi": self.cells_oi, "genes_oi": self.genes_oi}

    def filter_genes(self, genes):
        genes_oi = self.genes_oi[genes[self.genes_oi]]

        return Minibatch(
            cells_oi=self.cells_oi,
            genes_oi=genes_oi,
            phase=self.phase,
        )

    def to(self, device):
        self.device = device

    @property
    def genes_oi_torch(self):
        return torch.from_numpy(self.genes_oi).to(self.device)

    @property
    def cells_oi_torch(self):
        return torch.from_numpy(self.genes_oi).to(self.device)

    @property
    def n_cells(self):
        return len(self.cells_oi)

    @property
    def n_genes(self):
        return len(self.genes_oi)


class Minibatcher:
    def __init__(
        self,
        cells,
        genes,
        n_cells_step,
        n_genes_step,
        use_all_cells=False,
        use_all_genes=True,
        permute_cells=True,
        permute_genes=False,
    ):
        self.cells = cells
        if not isinstance(genes, np.ndarray):
            genes = np.array(genes)
        self.genes = genes
        self.n_genes = len(genes)
        self.n_cells_step = n_cells_step
        self.n_genes_step = n_genes_step

        self.permute_cells = permute_cells
        self.permute_genes = permute_genes

        self.use_all_cells = use_all_cells or len(cells) < n_cells_step
        self.use_all_genes = use_all_genes or len(genes) < n_genes_step

        self.cellxgene_batch_size = n_cells_step * n_genes_step

        # calculate length
        n_cells = len(cells)
        n_genes = len(genes)
        if self.use_all_cells:
            n_cell_bins = math.ceil(n_cells / n_cells_step)
        else:
            n_cell_bins = math.floor(n_cells / n_cells_step)
        if self.use_all_genes:
            n_gene_bins = math.ceil(n_genes / n_genes_step)
        else:
            n_gene_bins = math.floor(n_genes / n_genes_step)
        self.length = n_cell_bins * n_gene_bins

        self.i = 0

    def __len__(self):
        return self.length

    def __iter__(self):
        self.rg = np.random.RandomState(self.i)

        if self.permute_cells:
            cells = self.rg.permutation(self.cells)
        else:
            cells = self.cells
        if self.permute_genes:
            genes = self.rg.permutation(self.genes)
        else:
            genes = self.genes

        gene_cuts = [*np.arange(0, len(genes), step=self.n_genes_step)]
        if self.use_all_genes:
            gene_cuts.append(len(genes))
        gene_bins = [genes[a:b] for a, b in zip(gene_cuts[:-1], gene_cuts[1:])]

        cell_cuts = [*np.arange(0, len(cells), step=self.n_cells_step)]
        if self.use_all_cells:
            cell_cuts.append(len(cells))
        cell_bins = [cells[a:b] for a, b in zip(cell_cuts[:-1], cell_cuts[1:])]

        for cells_oi, genes_oi in itertools.product(cell_bins, gene_bins):
            yield Minibatch(cells_oi=cells_oi, genes_oi=genes_oi)

        self.i += 1



import torch
import numpy as np
import pyximport

pyximport.install(
    reload_support=True,
    language_level=3,
    setup_args=dict(include_dirs=[np.get_include()]),
)
from . import fragments_helpers
import dataclasses
from functools import cached_property

import chromatinhd.data.fragments


@dataclasses.dataclass
class Result:
    coordinates: torch.Tensor
    local_cellxgene_ix: torch.Tensor
    genemapping: torch.Tensor
    n_fragments: int
    cells_oi: np.ndarray
    genes_oi: np.ndarray
    window: np.ndarray
    n_total_genes: int

    @property
    def n_cells(self):
        return len(self.cells_oi)

    @property
    def n_genes(self):
        return len(self.genes_oi)

    def to(self, device):
        for field_name, field in self.__dataclass_fields__.items():
            if field.type is torch.Tensor:
                self.__setattr__(
                    field_name, self.__getattribute__(field_name).to(device)
                )
        return self

    @property
    def local_gene_ix(self):
        return self.local_cellxgene_ix % self.n_genes

    @property
    def local_cell_ix(self):
        return torch.div(self.local_cellxgene_ix, self.n_genes, rounding_mode="floor")

    @property
    def genes_oi_torch(self):
        return torch.from_numpy(self.genes_oi).to(self.coordinates.device)

    @property
    def cells_oi_torch(self):
        return torch.from_numpy(self.cells_oi).to(self.coordinates.device)

    def filter_fragments(self, fragments_oi):
        assert len(fragments_oi) == self.n_fragments
        return Result(
            coordinates=self.coordinates[fragments_oi],
            local_cellxgene_ix=self.local_cellxgene_ix[fragments_oi],
            genemapping=self.genemapping[fragments_oi],
            n_fragments=fragments_oi.sum(),
            cells_oi=self.cells_oi,
            genes_oi=self.genes_oi,
            window=self.window,
            n_total_genes=self.n_total_genes,
        )


def cell_gene_to_cellxgene(cells_oi, genes_oi, n_genes):
    return (cells_oi[:, None] * n_genes + genes_oi).flatten()


class Fragments:
    cellxgene_batch_size: int

    preloaded = False

    out_coordinates: torch.Tensor
    out_genemapping: torch.Tensor
    out_local_cellxgene_ix: torch.Tensor

    n_genes: int

    def __init__(
        self,
        fragments: chromatinhd.data.fragments.Fragments,
        cellxgene_batch_size: int,
        n_fragment_per_cellxgene: int = None,
    ):
        self.cellxgene_batch_size = cellxgene_batch_size

        # store auxilliary information
        window = fragments.regions.window
        self.window = window
        self.window_width = window[1] - window[0]

        # store fragment data
        self.cellxgene_indptr = fragments.cellxgene_indptr.numpy()
        self.coordinates = fragments.coordinates.numpy()
        self.genemapping = fragments.genemapping.numpy()

        # create buffers for coordinates
        if n_fragment_per_cellxgene is None:
            n_fragment_per_cellxgene = fragments.estimate_fragment_per_cellxgene()
        fragment_buffer_size = n_fragment_per_cellxgene * cellxgene_batch_size
        self.fragment_buffer_size = fragment_buffer_size

        self.n_genes = fragments.n_genes

    def preload(self):
        self.out_coordinates = torch.from_numpy(
            np.zeros((self.fragment_buffer_size, 2), dtype=np.int64)
        )  # .pin_memory()
        self.out_genemapping = torch.from_numpy(
            np.zeros(self.fragment_buffer_size, dtype=np.int64)
        )  # .pin_memory()
        self.out_local_cellxgene_ix = torch.from_numpy(
            np.zeros(self.fragment_buffer_size, dtype=np.int64)
        )  # .pin_memory()

        self.preloaded = True

    def load(self, minibatch):
        if not self.preloaded:
            self.preload()

        minibatch.cellxgene_oi = cell_gene_to_cellxgene(
            minibatch.cells_oi, minibatch.genes_oi, self.n_genes
        )

        assert len(minibatch.cellxgene_oi) <= self.cellxgene_batch_size, (
            len(minibatch.cellxgene_oi),
            self.cellxgene_batch_size,
        )
        n_fragments = fragments_helpers.extract_fragments(
            minibatch.cellxgene_oi,
            self.cellxgene_indptr,
            self.coordinates,
            self.genemapping,
            self.out_coordinates.numpy(),
            self.out_genemapping.numpy(),
            self.out_local_cellxgene_ix.numpy(),
        )
        if n_fragments > self.fragment_buffer_size:
            raise ValueError("n_fragments is too large for the current buffer size")

        if n_fragments == 0:
            n_fragments = 1
        self.out_coordinates.resize_((n_fragments, 2))
        self.out_genemapping.resize_((n_fragments))
        self.out_local_cellxgene_ix.resize_((n_fragments))

        return Result(
            local_cellxgene_ix=self.out_local_cellxgene_ix,
            coordinates=self.out_coordinates,
            n_fragments=n_fragments,
            genemapping=self.out_genemapping,
            window=self.window,
            n_total_genes=self.n_genes,
            **minibatch.items(),
        )
